module Markdown where

import Lucid
import Lucid.Base (makeElement)
import Data.Text (pack)
import Text.Pandoc.Builder
import Control.Monad (foldM)
import Data.Functor ((<&>))
import Data.Bifoldable (Bifoldable, bifoldMap, bifoldrM)
import Control.Monad.State.Strict
import Control.Monad.Identity

data MDState = MDState 
  { stNoteCount :: Int       -- count of notes that HAVE been printed
  , stNotes     :: [Html ()] -- notes that haven't be printed yet
  , stLevel     :: Int       -- block depth
  }

defaultMDState :: MDState
defaultMDState = MDState
  { stNoteCount = 0
  , stNotes     = []
  , stLevel     = 0
  }

render :: Pandoc -> Html ()
render (Pandoc meta blocks) =
  fst $ runIdentity $ runStateT (blocksToHtml blocks) defaultMDState

inlineToHtml :: Monad m => Inline -> StateT MDState m (Html ())
inlineToHtml (Str str) = pure $ toHtml str
inlineToHtml (Emph ins) = em_ <$> inlinesToHtml ins
inlineToHtml (Underline ins) = makeElement "u" <$> inlinesToHtml ins
inlineToHtml (Strong ins) = strong_ <$> inlinesToHtml ins
inlineToHtml (Strikeout ins) = makeElement "s" <$> inlinesToHtml ins
inlineToHtml (Superscript ins) = sup_ <$> inlinesToHtml ins
inlineToHtml (Subscript ins) = sub_ <$> inlinesToHtml ins
inlineToHtml (SmallCaps ins) = span_ [class_ "sc"] <$> inlinesToHtml ins
inlineToHtml (Quoted typ ins) =
  let (l, r) = case typ of
                 SingleQuote -> ("‘", "’")
                 DoubleQuote -> ("“", "”")
  in pure l >> inlinesToHtml ins >> pure r
inlineToHtml (Cite cits ins) = undefined -- TODO: citations
inlineToHtml (Code attrs str) = pure $ code_ $ toHtml str
inlineToHtml Space = pure " "
inlineToHtml SoftBreak = pure "\n"
inlineToHtml LineBreak = pure $ br_ []
inlineToHtml (Math _ math) = pure $ span_ [class_ "math"] $ toHtml math -- TODO: compile-time katex math
inlineToHtml (RawInline f str) | f == "html" = pure $ toHtmlRaw str
inlineToHtml (RawInline f str) = pure mempty
inlineToHtml (Link attrs ins (src, title)) = a_ [href_ src, title_ title] <$> inlinesToHtml ins
inlineToHtml (Image attrs ins (src, title)) = pure $ img_ [href_ src, title_ title] -- TODO: handle image alt
inlineToHtml (Note blocks) = do
  MDState {..} <- get
  let currentIdx = stNoteCount + length stNotes + 1
  note <- blocksToHtml blocks
  modify \s -> s { stNotes = note : stNotes }
  pure $ a_ [href_ $ "#note" <> pack (show currentIdx)] $ "[" <> toHtml (show currentIdx) <> "]"
inlineToHtml (Span attrs ins) = span_ <$> inlinesToHtml ins

inlinesToHtml :: Monad m => [Inline] -> StateT MDState m (Html ())
inlinesToHtml = foldMapM inlineToHtml

blockToHtml :: Monad m => Block -> StateT MDState m (Html ())
blockToHtml (Plain ins) = pre_ <$> inlinesToHtml ins
blockToHtml (Para ins) = p_ <$> inlinesToHtml ins
blockToHtml (LineBlock ins) = undefined
blockToHtml (CodeBlock attrs code) = pure $ pre_ $ code_ $ toHtmlRaw code -- TODO: highlighting
blockToHtml (RawBlock f str) | f == "html" = pure $ toHtmlRaw str
blockToHtml (RawBlock f str) = pure mempty
blockToHtml (BlockQuote blocks) = blockquote_ <$> blocksToHtml blocks
blockToHtml (OrderedList attrs items) = ol_ <$> foldMapM (fmap li_ . blocksToHtml) items -- TODO: handle attrs
blockToHtml (BulletList items) = ul_ <$> foldMapM (fmap li_ . blocksToHtml) items
blockToHtml (DefinitionList items) =
  dl_ <$> foldMapM (bifoldMapM inlinesToHtml (foldMapM (fmap dd_ . blocksToHtml))) items
blockToHtml (Header k attrs ins) = h_ k <$> inlinesToHtml ins -- TODO: handle attrs (especially ID)
  where h_ :: Applicative m => Int -> HtmlT m a -> HtmlT m a
        h_ 1 = h1_
        h_ 2 = h2_
        h_ 3 = h3_
        h_ 4 = h4_
        h_ 5 = h5_
        h_ _ = h6_
blockToHtml HorizontalRule = pure $ hr_ []
blockToHtml (Table attrs caption _ head rows foot) = pure mempty
blockToHtml (Div attrs blocks) = pure mempty
blockToHtml Null = pure mempty

blocksToHtml :: Monad m => [Block] -> StateT MDState m (Html ())
blocksToHtml = fmap mconcat . mapM renderBlock
  where renderBlock :: Monad m => Block -> StateT MDState m (Html ())
        renderBlock block = do
          modify \s -> s { stLevel = stLevel s + 1 }
          html <- blockToHtml block
          MDState {..} <- get
          let printingNotes = not (null stNotes) && stLevel == 1

          let notes = when printingNotes (ol_ [class_ "footnotes", start_ $ pack (show $ stNoteCount + 1)] $ foldMap li_ (reverse stNotes))
          when printingNotes $ modify \s -> s {stNoteCount = stNoteCount + length stNotes, stNotes = [] }

          modify \s -> s { stLevel = stLevel - 1 }

          pure $ html <> notes

foldMapM :: (Monad m, Monoid w, Foldable t) => (a -> m w) -> t a -> m w
foldMapM f = foldM (\acc a -> do
                     w <- f a
                     return $! mappend acc w)
                   mempty

bifoldMapM :: (Monad m, Monoid w, Bifoldable t) => (a -> m w) -> (b -> m w) -> t a b -> m w
bifoldMapM f g = bifoldrM (\a b -> f a <&> (`mappend` b)) (\a b -> g a <&> (`mappend` b)) mempty
